@@ -7,6 +7,7 @@ use constant IV_SIZE => eval 'require Config;$Config::Config{ivsize}';
 
 my %COLLECTION_RE = (pipes => qr{\|}, csv => qr{,}, ssv => qr{\s}, tsv => qr{\t});
 
+has _api_spec => sub { die '_api_spec() cannot be built' };
 has _json_validator => sub { state $v = JSON::Validator->new; };
 
 sub validate_input {
@@ -27,6 +28,11 @@ sub validate_request {
       $value = $self->_extract_request_parameter($c, $in);
       $exists = length $value;
     }
+    elsif ($in eq 'formData' and $type eq 'file') {
+      $value  = $c->req->upload($name);
+      $exists = $value ? 1 : 0;
+      $value  = $value ? $value->slurp : undef;
+    }
     else {
       $value  = $cache{$in} ||= $self->_extract_request_parameter($c, $in);
       $exists = exists $value->{$name};
@@ -160,14 +166,22 @@ sub _validate_type_array {
   return $self->SUPER::_validate_type_array(@_[1, 2, 3]);
 }
 
-# always valid
-sub _validate_type_file { }
+sub _validate_type_file {
+  my ($self, $data, $path, $schema) = @_;
+
+  if ($schema->{required} and (not defined $data or not length $data)) {
+    return JSON::Validator::E($path => "Missing property.");
+  }
+
+  return;
+}
 
 sub _validate_type_object {
   return shift->SUPER::_validate_type_object(@_) unless $_[0]->{validate_input};
 
   my ($self, $data, $path, $schema) = @_;
   my $properties = $schema->{properties} || {};
+  my $discriminator = $schema->{discriminator};
   my (%ro, @e);
 
   for my $p (keys %$properties) {
@@ -176,6 +190,15 @@ sub _validate_type_object {
     $ro{$p} = 1;
   }
 
+  if ($discriminator and !$self->{inside_discriminator}) {
+    my $name = $data->{$discriminator}
+      or return JSON::Validator::E($path, "Discriminator $discriminator has no value.");
+    my $dschema = $self->_api_spec->get("/definitions/$name")
+      or return JSON::Validator::E($path, "No definition for discriminator $name.");
+    local $self->{inside_discriminator} = 1;    # prevent recursion
+    return $self->_validate($data, $path, $dschema);
+  }
+
   local $schema->{required} = [grep { !$ro{$_} } @{$schema->{required} || []}];
 
   return @e, $self->SUPER::_validate_type_object($data, $path, $schema);
